# Use case to update a register
# @note
#   - You can define methods `filters` and `search` to change the target entries of the register
#   - You need to define the `process_ooze` method
# This case expects `options[:source][:register_id]`
class Eco::API::UseCases::OozeSamples::RegisterUpdateCase < Eco::API::UseCases::OozeSamples::OozeBaseCase

  class << self
    # @return [Integer] the number of pages to be processed in each batch
    def batch_size(size = nil)
      @batch_size ||= 25
      return @batch_size unless size
      @batch_size = size
    end
  end

  name "register-update-case"
  type :other

  def main(session, options, usecase, &block)
    super(session, options, usecase) do
      with_each_entry do |ooze|
        process_ooze(&block)
      end
    end
  end

  def process_ooze(ooze = target)
    raise "You need to define this method"
  end

  private

  def before_loading_new_target(ooze_id)
    if pending = queue_shift(ooze_id)
      update_ooze(pending)
    end
  end

  def new_target(object, warn_pending_changes: false)
    enqueue(object)
    super(object, warn_pending_changes: warn_pending_changes) do |pending|
      #enqueue(pending)
    end
  end

  def with_each_entry
    batched_search_results do |page_results|
      page_results.each do |page_result|
        if pending = queue_shift(page_result.id)
          if dirty?(pending)
            msg  = "Inconsistent search results. "
            msg << "Launching update on '#{object_reference(pending)}' to be able to queue it back"
            console.warn msg
            update_ooze(pending)
          end
        end
        if ooz = ooze(page_result.id)
          yield(ooz)
        end
      end
      update_oozes
    end
  end

  def enqueue(object)
    return unless object && object.respond_to?(:id)
    unless object.is_a?(Ecoportal::API::V2::Page) or object.is_a?(Ecoportal::API::V2::Pages::PageStage)
      raise "Queuing is just for entries. Expecting Page or PageStage. Given: #{object.class}"
    end
    if elem = queue_get(object.id)
      msg = "Something is wrong. Native case 'RegisterUpdateCase' is "
      if (elem != object) && dirty?(elem)
        msg << "trying to queue different objects with same page id:\n"
        msg << " • already queued (changes will go): #{object_reference(elem)}\n"
        msg << " • tried to queue (lost changes): #{object_reference(object)}"
        logger.warn msg
      end
    else
      batch_queue << object unless batch_queue.include?(object)
    end
  end

  def queue_shift(id = nil)
    if id
      batch_queue.delete(queue_get(id))
    else
      batch_queue.shift
    end
  end

  def queue_get(id)
    batch_queue.find {|e| e.id == id}
  end

  def batch_queue
    @batch_queue ||= []
  end

  def batched_search_results
    raise "Missing block. It yields in slices of #{self.class.batch_size} results" unless block_given?
    results_preview
    results = []
    apiv2.registers.search(register_id, search_options) do |page_result|
      results << page_result
      if results.length >= self.class.batch_size
        yield(results)
        results = []
      end
    end
    yield(results) unless results.empty?
  end

  def update_oozes(batched_oozes = batch_queue)
    batched_oozes.each do |ooze|
      update_ooze(ooze)
    end
    batched_oozes.clear
  end

  def backup_patch!(ooze = target)
    unless patch = (patch_doc(ooze) || {})["page"]
      logger.info "No changes to update for #{object_reference(ooze)}."
      return
    end
  end

  def results_preview
    apiv2.registers.search(register_id, search_options.merge(only_first: true)).tap do |search_results|
      str_results = "Total target entries: #{search_results.total} (out of #{search_results.total_before_filtering})"
      session.prompt_user("Do you want to proceed (y/N):", explanation: str_results, default: "N", timeout: 10) do |res|
        if res.upcase.start_with?("N")
          puts "..."
          logger.info "Aborting script..."
          exit(0)
        end
      end
    end
  end

  def search_options
    @search_options ||= {}.tap do |opts|
      opts.merge!(sort: "created_at")
      opts.merge!(dir:  "asc")
      opts.merge!(query:   conf_filters) if conf_search
      opts.merge!(filters: conf_filters)
    end
  end

  def conf_filters
    return filters if self.respond_to?(:filters)
    []
  end

  def conf_search
    return search if self.respond_to?(:search)
  end

  def register_id
    options.dig(:source, :register_id)
  end

end
